#define version "0.7b"

// wcrush by CS - a 64bit Windows compatible replacement for Taboo's levelpacker.

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>

#define maxsize 65537
#define endmark 511
#define maxref 255

// working buffers + some extra bytes, just in case:

uint8_t inbuf[maxsize + 1]; 
uint8_t outbuf[maxsize + 350]; 

// used everywhere, so they are global vars:

uint32_t srcidx, tgtidx, ctrlidx, uncmp, ctrlbit;

void putbit (uint32_t bit);
void storelenbits (uint32_t l, uint32_t stopbit);
void storeoffset (uint32_t offset, uint32_t table);
void storeref(uint32_t offset, uint32_t size, uint32_t table);
void copybytes ();
uint32_t packed (uint32_t end, int speed);

int main(int argc, char* argv[])
{
    printf("\ncrush for windows v%s by CS\n\n", version);

// we need at least a speed from 0 to 6 and an inputfilename as parameters:

    int speed = 7;
    if (argv[1] != NULL)
        speed = atoi(argv[1]);
    if (argc < 3 || speed < 0 || speed > 6) 
    {
        printf ("usage: wcrush <speed 0..6> <source name> [target name]\n");
        exit (EXIT_FAILURE);
    }
    char dftname[] = "crushed";
    if (argc < 4)
        argv[3] = dftname;
    printf ("trying to crush %s...\n\n", argv[2]);
    FILE * infile;
    infile = fopen (argv[2], "rb");
    
// inputfile must be > 2 bytes and < 65538 bytes in size and should be openable :-)
    
    if (infile == NULL) 
    {
        printf ("can't open %s!\n", argv[2]);
        exit (EXIT_FAILURE);
    }
    fseek (infile, 0, SEEK_END);
    long ifsize = ftell(infile);
    rewind (infile);
    if (ifsize > maxsize)
    {
        printf ("file too large!\n");
        exit (EXIT_FAILURE);
    }
    if (ifsize < 3)
    {
        printf("file too small!\n");
        exit(EXIT_FAILURE);
    }

    fread (inbuf, 1, ifsize, infile);
    fclose (infile);
    printf ("size of %s: %ld bytes.\n", argv[2], ifsize);

// pack it and throw error if result too large, else copy load addr and try to write it to disk:
    
    uint32_t ofsize = packed (ifsize, speed);
    if (ofsize > maxsize)
    {
        printf("couldn't pack file!\n");
        exit(EXIT_FAILURE);
    }
    outbuf[0] = inbuf[0]; 
    outbuf[1] = inbuf[1];
    printf("crushed size: %d bytes, %ld%% left\n", ofsize, 100 * ofsize / ifsize);
    FILE * outfile;
    outfile = fopen(argv[3], "wb");
    if (outfile == NULL) 
    {
        printf ("can't write %s!\n", argv[3]);
        exit (EXIT_FAILURE);
    }
    fwrite (outbuf, 1, ofsize, outfile);
    fclose (outfile);
    return EXIT_SUCCESS;
}


// pack the file in the format the unpacker expects, return the size of the packed file
// or 0 if it got larger than the input file or some other terrible things happened:

uint32_t packed(uint32_t end, int speed)
{
// maximum distance to search back for 3 byte and 2 byte long references, depending on speed:    
    
    const uint16_t longmax[7] = { 115, 343, 679, 1359, 2383, 4687 , 9359 };
    const uint16_t shortmax[7] = { 115, 343, 679, 1359, 1359, 1359, 1359 };

// reference search related variables:

    uint32_t refidx, distance, bestoffs, bestlen, matchlen, max, table,
             maxdistlong, maxdistshort, maxdistmax;

// construct a lookup table that holds the next earlier position of a 2 byte sequence matching
// the one at the current position in the inputfile. 0 means the sequence wasn't there yet:

    uint32_t previous[maxsize + 1];
    uint16_t last[65536] = {0};

    for (refidx = 2; refidx < end; ++refidx)
    {
        previous[refidx] = last[inbuf[refidx] + inbuf[refidx+1] * 256];
        last[inbuf[refidx] + inbuf[refidx+1] * 256] = refidx;
    }

// set some speed dependent settings: table is used to pick the proper amount of bits to
// store the distance of a reference to the current position, what in turn affects
// maxdistshort, the largest possible distance to the current position for a 2 byte long 
// reference, maxdistlong, the same for a 3 byte long reference and maxdistmax, the same for
// the longest possible reference:

    table = speed * 8;
    maxdistshort = shortmax[speed];
    maxdistlong = longmax[speed];
    maxdistmax = maxdistlong + maxref;

// srcidx = position of first byte in input buffer after load address, 
// ctrlidx = first controlbyte pos in output buffer,
// tgtidx = first databyte pos in output buffer (must not overwrite controlbyte)
// uncmp = length of unpackable byte sequence (copied unchanged from input- to output buffer)
// ctrlbit = first control bit to write into controlbyte (bit#7):

    srcidx = 2;
    ctrlidx = 2;
    tgtidx = 3;
    uncmp = 0;
    ctrlbit = 0x80;

// the (very naively designed) compression loop: 

    while (srcidx < end && tgtidx <= maxsize)
    {

// start search for matching byte sequence at current position, assuming no match has been found yet:

        bestlen = 0;        
        refidx = previous[srcidx];

// reference must be at least 2 bytes away as overlapping with destination isn't allowed:

        if (srcidx - refidx < 2)
            refidx = previous[refidx];

// find the next earlier byte of the same value within the maximum possible search range and
// see if and how many following bytes also match the following bytes at the current position:

        while (refidx && (distance=srcidx-refidx) <= maxdistmax && bestlen < maxref)
        {

// maximum of bytes to compare: theoritical limit = 255, practical limit = distance between matches,
// and near the end of input file distance between input pos and end to avoid adding garbage:

            matchlen = 2;
            if (distance < maxref)
                max = distance;
            else
                max = maxref;
            if (srcidx + max > end)
                max = end - srcidx;

// count number of matching bytes, and remember their position if the matching sequence is
// longer than a previously found one:

            while (matchlen < max && inbuf[refidx + matchlen] == inbuf[srcidx + matchlen])
                ++matchlen;

            if (matchlen > bestlen && distance - matchlen <= maxdistlong)
            {
                bestlen = matchlen;
                bestoffs = distance;
            }
            refidx=previous[refidx];
        }

// if a match of at least 2 bytes has been found, write a reference to that match into the
// output buffer, else count the current byte as unpackable and check the next byte:

        bestoffs -= bestlen;
        if (bestlen > 2 || (bestlen == 2 && bestoffs <= maxdistshort))
            storeref(bestoffs, bestlen, table);
        else
        {
            ++uncmp;
            ++srcidx;
        }
    }

// end of compression loop: if there are unpackable bytes remaining copy them to the 
// output buffer now, else add a "reference coming up"-flag to the control bit stream:

    if (uncmp)  
        copybytes();
    else
        putbit(1); 

// add endmark (= reference length of 511 bytes w/o final stopbit):

    putbit(1);
    storelenbits(endmark, 0);

// fill up the final control byte with 0 bits to ensure it is written to the output buffer:

    while (ctrlbit)
        putbit(0);
    return tgtidx;
    
}


// encode a 2 or more bytes long repetition into output data:

void storeref(uint32_t offset, uint32_t size, uint32_t table)
{

// If unpackable bytes have been found copy them to the output buffer first, 
// else add a "repetition ahead" flagbit to the control bitstream.

    if (uncmp) 
        copybytes ();
    else
        putbit(1); 

// encode reference length: 2 byte long matches only need a 0 bit, 3 byte long ones a stopbit,
// anything longer a proper stopbit, lengthbit... sequence. Also note that 2 byte references
// use different table entries to encode the offset:

    if (size == 2)
    {
        putbit(0);
        storeoffset(offset, table + 4);
    }
    else
    {
        putbit(1);
        if (size == 3)
            putbit(1);
        else
            storelenbits(size - 2, 1);
        storeoffset(offset, table);
    }

// do not copy the repeated bytes to the output buffer, continue search for matches after them:

    srcidx += size;
}


// encode the relative position of a reference. this is a bit complicated as the depacker first
// expects 2 control bits telling it how many "bitchunks" will follow. each chunk has a specific,
// speed dependent number of bits containg the offset (or parts thereof, in case multiple chunks
// are used). The depacker expects the offset to point to the END of the reference to extend the
// search range (since the reference length is known the start address can be easily calculated 
// by adding it to the offset):

void storeoffset(uint32_t offset, uint32_t table)
{
    
// number of bits per chunk, values 0-3 of each row are used for 3+ byte long references, 
// 4-7 for 2 byte long ones:

    const uint8_t packbits[56] = {
        2, 2, 1, 1, 2, 2, 1, 1,  // worst compression (speed 0)
        3, 1, 2, 2, 3, 1, 2, 2,
        3, 2, 2, 2, 3, 2, 2, 2,
        4, 2, 2, 2, 4, 2, 2, 2,
        4, 2, 2, 3, 4, 2, 2, 2,
        4, 2, 3, 3, 4, 2, 2, 2,
        4, 3, 3, 3, 4, 2, 2, 2 }; // best compression (speed 6)

// total number of bits used so far = 0,
// largest offset that can be stored with current amount of bits = 0,
// currently used bitchunk = -1 (due to preincrement, see below)
// current compensation value = 0 (to compensate for depacker adding 1 to partial results):

    uint32_t bits = 0;
    uint32_t largest = 0;
    int32_t chunk = -1; 
    uint32_t subst = 0;

// keep adding bitchunks until we have enough bits in total to encode the offset, and whenever
// we need an additional chunk, update the compensation value, too: 

    while (offset >= largest)
    {
        subst = largest;  
        bits += packbits[++chunk + table];
        largest += (1 << bits); 
    }

// the depacker adds 1 to the partially calculated offset whenever another bitchunk is needed, 
// and the result is leftshifted by the amount of bits to follow to extend the search range, so
// we have to substract the compensation value to get the proper value to store:

    offset -= subst; 

// store number of chunks used:

    putbit(chunk & 2);
    putbit(chunk & 1);

// store the offset bits, starting with the most significant one:

    bits = 1 << bits;
    while (bits /= 2)
        putbit(offset&bits);
}


// copy an incompressable byte sequence from input to output data:

void copybytes()
{

// flag run of uncompressed bytes ahead and store its length:

    putbit(0); 
    storelenbits(uncmp, 1);

// copy those bytes to output data (and reset # of uncompressed bytes to 0 in the process):

    while (uncmp)
        outbuf[tgtidx++] = inbuf[srcidx - uncmp--];
}


// this will create the stopbit/bit sequence which the depacker uses to calculate lengths.
// the final stopbit is an extra parameter as the end mark doesn't need it:

void storelenbits(uint32_t len, uint32_t stopbit)
{
    
// figure out the most significant bit of len:

    uint32_t msbval = 1;
    uint32_t tmp = len;
    while (tmp /= 2)
        msbval *= 2; 

// remove it as the unpacker will restore it automatically:

    tmp = len - msbval;

// write a sequence of "no-stop"-bits each followed by a length bit, msb first:

    while (msbval /= 2)
    {
        putbit(0);
        putbit(tmp & msbval);
    }

    if (stopbit)
        putbit(1);
}


// collect control bits and store them at the next unused outbuf position.
// IMPORTANT: The correct order of operations is to not add the control byte
// to output data when it is full but when it is full AND another bit should
// be added! The difference is subtle and may or may not derail the depacker,
// depending on the payload, so when errors occur after depacking in a rather
// random fashion this routine is likely the culprit:

void putbit(uint32_t bit)
{

// if the current control byte is full, switch to the next unused byte in the output buffer, 
// update the next databyte position so it won't overwrite the new control byte, clear all
// bits and reset the countrol bit mask/counter to the most significant bit:

    if (!ctrlbit)
    {
        ctrlidx = tgtidx++;
        outbuf[ctrlidx] = 0;
        ctrlbit = 0x80;
    }

// any input value != 0 will set the current bit, a value == 0 leaves it clear:

    if (bit)
        outbuf[ctrlidx] |= ctrlbit;
    ctrlbit /= 2;
}
